﻿using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Digests;
using Org.BouncyCastle.Crypto.Encodings;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Modes;
using Org.BouncyCastle.Crypto.Parameters;
using System;
using System.IO;
using System.Text;
using System.Xml;

namespace Saml2Code
{

    public class CipherUtils
    {
        /// <summary>
        /// Does first asymmetric decrypt
        /// </summary>
        /// <param name="xml"> The xml encripted </param>
        /// <returns> Symmetric key decrypted </returns>
        public static byte[] AsymmetricDecrypt(XmlDocument xml)
        {
            string data = xml.GetElementsByTagName("xenc:CipherValue")[0].InnerText;

            IAsymmetricBlockCipher cipher = new OaepEncoding(new RsaEngine(), new Sha1Digest());

            AsymmetricCipherKeyPair key = (AsymmetricCipherKeyPair) Utils.CreateKey(1);

            RsaKeyParameters privateKey = (RsaKeyParameters)key.Private;

            cipher.Init(false, privateKey);

            byte[] cipheredBytes = Convert.FromBase64String(data);

            byte[] deciphered = cipher.ProcessBlock(cipheredBytes, 0, cipheredBytes.Length);

            return deciphered;
        }

        /// <summary>
        /// Decrypts a xml. Needs the rest of methods from this class
        /// </summary>
        /// <param name="xml"> XML that to be decrypted </param>
        /// <param name="nonSecretPayloadLength"> Must be 0 </param>
        /// <returns> String with xml decripted </returns>
        public static string Decrypt(XmlDocument xml, int nonSecretPayloadLength = 0)
        {
            var key = AsymmetricDecrypt(xml);

            byte[] data = Convert.FromBase64String(xml.GetElementsByTagName("xenc:CipherValue")[1].InnerText);

            var plaintext = DecryptWithKey(data, key, nonSecretPayloadLength);

            return Encoding.ASCII.GetString(plaintext);
        }

        /// <summary>
        /// Makes a symmetric decrypt (this will be the second step. 
        /// Needs to do a asymmetric decrypt in the first step)
        /// </summary>
        /// <param name="encryptedMessage"> The encripted data in byte array</param>
        /// <param name="key"> the key from asymmetric method</param>
        /// <param name="nonSecretPayloadLength"> Must be 0</param>
        /// <returns> byte array with attributes decripted </returns>
        private static byte[] DecryptWithKey(byte[] encryptedMessage, byte[] key, 
            int nonSecretPayloadLength = 0)
        {
            if (key == null || key.Length != 256 / 8)
            {
                throw new ArgumentException(String.Format("Key needs to be {0} bit!", 256), "key");
            }

            if (encryptedMessage == null || encryptedMessage.Length == 0)
            {
                throw new ArgumentException("Encrypted Message Required!", "encryptedMessage");
            }

            using (var cipherStream = new MemoryStream(encryptedMessage))
            using (var cipherReader = new BinaryReader(cipherStream))
            {
                var nonce = cipherReader.ReadBytes(96 / 8);

                var cipher = new GcmBlockCipher(new AesFastEngine());
                var parameters = new AeadParameters(new KeyParameter(key), 128, nonce, null);
                cipher.Init(false, parameters);

                var cipherText = cipherReader.ReadBytes(encryptedMessage.Length - nonSecretPayloadLength - nonce.Length);
                var plainText = new byte[cipher.GetOutputSize(cipherText.Length)];

                try
                {
                    var len = cipher.ProcessBytes(cipherText, 0, cipherText.Length, plainText, 0);
                    cipher.DoFinal(plainText, len);
                }
                catch (InvalidCipherTextException)
                {
                    //Return null if it doesn't authenticate
                    return null;
                }
                return plainText;
            }
        }
    }
}

