﻿using System;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Security.Cryptography.Xml;
using System.Xml;


namespace Saml2Code
{
    public class Sign
    {
        /// <summary>
        /// method will sign the metadata XML or the saml2 request 
        /// </summary>
        /// <param name="Key"> RSACryptoServiceProvider.</param>
        /// <param name="cert"> your certicate.p12 and your password.</param>
        /// <param name="doc"> xml that we want to sign </param>
        /// <param name="UriNode"> Node to sign</param>
        /// <returns> XML signed </returns>
        public static XmlDocument SignXmlFile(RSA Key, X509Certificate2 cert, XmlDocument doc, string UriNode)
        {
            try
            {
                doc.PreserveWhitespace = true;
                
                SignedXml signedXml = new SignedXml(doc);

                signedXml.SigningKey = Key;

                signedXml.SignedInfo.SignatureMethod = Conf.get_sign_method();
                signedXml.SignedInfo.CanonicalizationMethod = Conf.get_canonicalization_method();

                KeyInfo keyInfo = new KeyInfo();
                KeyInfoX509Data keyInfoData = new KeyInfoX509Data(cert);
                keyInfo.AddClause(keyInfoData);
                signedXml.KeyInfo = keyInfo;

                Reference reference = new Reference();
                //If you want to sign only a node, put the UriNode
                if(UriNode != null) reference.Uri = "#" + UriNode;
                else reference.Uri = "";
                reference.DigestMethod = Conf.get_digest_method();

                XmlDsigEnvelopedSignatureTransform env = new XmlDsigEnvelopedSignatureTransform();
                reference.AddTransform(env);

                signedXml.AddReference(reference);

                signedXml.ComputeSignature();

                XmlElement xmlDigitalSignature = signedXml.GetXml();

                //doc.DocumentElement.AppendChild(doc.ImportNode(xmlDigitalSignature, true));
                doc.DocumentElement.InsertAfter(doc.ImportNode(xmlDigitalSignature, true), doc.GetElementsByTagName("saml2:Issuer")[0]);

                if (doc.FirstChild is XmlDeclaration)
                {
                    doc.RemoveChild(doc.FirstChild);
                }
                return doc;

            }
            catch (Exception e)
            {
                //LOG: " Error signing the xml: " + e
                return null;
            }
        }

        /// <summary>
        /// verifies the signature from IdP-Proxy metadata.
        /// </summary>
        /// <param name="cert"> cert .p12.</param>
        /// <param name="response"> XML that you want to verify.</param>
        /// <returns> if it was verified, will be true</returns>
        public static Boolean VerifyXmlFile(X509Certificate2 cert, XmlDocument response)
        {
            SignedXml signedXml2 = null;

            try
            {
                signedXml2 = new SignedXml(response);
            }
            catch (Exception e)
            {
                //LOG: "Error loading metadata XML file. Check if the name is right: " + e
            }

            try
            {
                XmlElement nodeList = (XmlElement)response.GetElementsByTagName("Signature")[0];

                if (nodeList == null)
                {
                    nodeList = (XmlElement)response.GetElementsByTagName("ds:Signature")[0];
                }

                signedXml2.LoadXml(nodeList);
            }
            catch (Exception e)
            {
                //LOG: " Error loading Signature from XML: " + e
            }

            bool isSigned = false;

            try
            {
                isSigned = signedXml2.CheckSignature(cert, true);
            }
            catch (Exception e)
            {
                //LOG: "Error cheking the signature from XML file: " + e
            }

            if (isSigned)
            {
                //LOG: "The XML signature is valid."
            }
            else
            {
                //LOG: "The XML signature is not valid."
            }

            return isSigned;
        }
        /// <summary>
        /// Verifies the signature from IdP-Proxy metadata.
        /// </summary>
        /// <param name="cert"> your cert .p12.</param>
        /// <param name="xmldoc"> XML that you want to verify.</param>
        /// <returns> if it was verified, will be true</returns>
        public static Boolean VerifyXmlFile(XmlDocument xmldoc, X509Certificate2 cert)
        {
            SignedXml signedXml2 = null;

            try
            {
                signedXml2 = new SignedXml(xmldoc);
            }
            catch (Exception e)
            {
                //LOG: " Error loading XML file. " + e
            }

            try
            {
                XmlElement nodeList = (XmlElement)xmldoc.GetElementsByTagName("Signature")[0];

                if (nodeList == null)
                {
                    nodeList = (XmlElement)xmldoc.GetElementsByTagName("ds:Signature")[0];
                }

                signedXml2.LoadXml(nodeList);
            }
            catch (Exception e)
            {
                //LOG: " Error loading Signature from XML: " + e
            }

            bool isSigned = false;

            try
            {
                isSigned = signedXml2.CheckSignature(cert, true);
            }
            catch (Exception e)
            {
                //LOG: " Error cheking the signature from XML file: " + e
            }

            if (isSigned)
            {
                //LOG: "The XML signature is valid."
            }
            else
            {
                //LOG: "The XML signature is not valid."
            }

            return isSigned;
        }
    }
}