﻿using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.OpenSsl;
using System;
using System.IO;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Web;
using System.Windows.Forms;
using System.Xml;

namespace Saml2Code
{
    public class Utils
    {
        /// <summary>
        /// Verifies if metadata from IdP-Proxy have the attributes necesary, 
        /// the signature is right, the cert is in the key store and the algorithm used in the
        /// signature is the same that SP uses
        /// </summary>
        /// <param name="xml">The xml to verify </param>
        /// <param name="logout">If it is a try of Log out</param>
        /// <returns>if response has been verified</returns>
        public static bool VerifyResponse(XmlDocument xml, Boolean logout)
        {
            return (VerifyCertInKeyStore(xml, logout) && VerifySignature(xml, true, logout));
        }

        /// <summary>
        /// Creates the certificate that SP will use 
        /// </summary>
        /// <param name="number">must be the number looking for your requirement (sign, encrypt...)</param>
        /// <returns>will return the certificate</returns>
        public static X509Certificate2 CreateCert(int number)
        {
            try
            {
                //RSACryptoServiceProvider Key = new RSACryptoServiceProvider();
                X509Certificate2 cert;

                switch (number)
                {
                    case 1:
                        //cert for encrypt xmls
                        String pem = new StreamReader(HttpContext.Current.Server.MapPath(Conf.get_dir_p12_encryption())).ReadToEnd();
                        var start = pem.IndexOf("-----BEGIN CERTIFICATE-----", StringComparison.Ordinal);
                        start += "-----BEGIN CERTIFICATE-----".Length;
                        var end = pem.IndexOf("-----END CERTIFICATE-----", start, StringComparison.Ordinal) - start;
                        return new X509Certificate2(Convert.FromBase64String(pem.Substring(start, end)));
                    case 2:
                        //cert for sign xmls
                        cert = new X509Certificate2(HttpContext.Current.Server.MapPath(Conf.get_dir_p12_signing()),
                            Conf.get_cert_password_sign());
                        break;
                    case 3:
                        string tumbprint = Conf.get_cert_thumbprint_sign();
                        cert = Clave.CertificateUtils.GetCertificateFromPersonalStore(tumbprint);
                        if (cert == null || !cert.HasPrivateKey)
                            throw new Exception("Certificate '" + tumbprint + "' not found at " +
                                "LocalMachine/My keystore or access to private key was denied. Certificate: " + cert);
                        break;
                    default:
                        cert = null;
                        break;
                }

                //var enhCsp = Key.CspKeyContainerInfo;
                //var cspparams = new CspParameters(enhCsp.ProviderType, enhCsp.ProviderName,
                //    ((RSACryptoServiceProvider)cert.PrivateKey).CspKeyContainerInfo.KeyContainerName);
                //Key = new RSACryptoServiceProvider(cspparams);

                //cert.PrivateKey = Key;

                return cert;
            }
            catch (Exception e)
            {
                //TODO
                return null;
            }
        }

        /// <summary>
        /// Creates RSA key that the SP will need to use 
        /// </summary>
        /// <param name="number">must be the number looking for your requirement (sign, encrypt...)</param>
        /// <returns>returns the cert</returns>
        public static Object CreateKey(int number)
        {

            //https://docs.microsoft.com/es-es/dotnet/standard/security/how-to-encrypt-xml-elements-with-asymmetric-keys
            CspParameters cspParams = new CspParameters();
            cspParams.KeyContainerName = "XML_ENC_RSA_KEY";

            RSACryptoServiceProvider Key;

            X509Certificate2 cert;

            switch (number)
            {
                case 1:
                    //cert for encryption
                    TextReader reader;
                    using (reader = File.OpenText(@HttpContext.Current.Server.MapPath(Conf.get_dir_p12_encryption())))
                        return (AsymmetricCipherKeyPair)new PemReader(
                            reader,
                            new DefaultPasswordFinder(Conf.get_cert_password_enc().ToCharArray())
                            ).ReadObject();
                case 2:
                    //cert for sign xmls
                    cert = new X509Certificate2(HttpContext.Current.Server.MapPath(Conf.get_dir_p12_signing()),
                        Conf.get_cert_password_sign());
                    break;
                case 3:
                    string tumbprint = Conf.get_cert_thumbprint_sign();
                    cert = Clave.CertificateUtils.GetCertificateFromPersonalStore(tumbprint);
                    if (cert == null || !cert.HasPrivateKey)
                        throw new Exception("Certificate '" + tumbprint + "' not found at " +
                            "LocalMachine/My keystore or access to private key was denied. Certificate: " + cert);
                    break;
                default:
                    cert = null;
                    break;
            }

            /*
            Key = new RSACryptoServiceProvider(cspParams);
           
            var enhCsp = Key.CspKeyContainerInfo;
            CspParameters cspparams = new CspParameters(enhCsp.ProviderType, enhCsp.ProviderName,
                ((RSACryptoServiceProvider)cert.PrivateKey).CspKeyContainerInfo.KeyContainerName);

            Key = new RSACryptoServiceProvider(cspparams);
            */

            const int PROV_RSA_AES = 24;
            CspKeyContainerInfo info = ((RSACryptoServiceProvider)cert.PrivateKey).CspKeyContainerInfo;
            CspParameters cspParameters = new CspParameters(PROV_RSA_AES)
            {
                KeyContainerName = info.KeyContainerName,
                KeyNumber = (int)info.KeyNumber,
                Flags = CspProviderFlags.UseExistingKey,
            };

            if (info.MachineKeyStore)
            {
                cspParameters.Flags |= CspProviderFlags.UseMachineKeyStore;
            }

            if (info.ProviderType == PROV_RSA_AES)
            {
                cspParameters.ProviderName = info.ProviderName;
            }
            Key = new RSACryptoServiceProvider(cspParameters);

            return Key;
        }

        /// <summary>
        /// Reads a xml, gets public key, creates the public cert 
        /// and optionally saves this in keystore
        /// </summary>
        /// <param name="xml">XML to verify</param>
        /// <param name="logout">If it is a logout try</param>
        /// <returns>if it was verified</returns>
        public static bool VerifyCertInKeyStore(XmlDocument xml, Boolean logout)
        {
            try
            {
                string publicKey = null;

                if (logout) publicKey = XMLData.ReadPublicKey(xml,
                    "/saml2p:LogoutResponse/ds:Signature/ds:KeyInfo/ds:X509Data/ds:X509Certificate", true);

                else publicKey = XMLData.ReadPublicKey(xml,
                    "/saml2p:Response/ds:Signature/ds:KeyInfo/ds:X509Data/ds:X509Certificate", true);

                X509Certificate2 cert = new X509Certificate2(Encoding.ASCII.GetBytes(publicKey));

                if (Keys.IsCertTrustworthy(cert))
                {
                    return true;
                }
                else
                {
                    if (MessageBox.Show("No se confía en el certificado " + cert.IssuerName.Name + " - Serial: " + cert.SerialNumber
                        + "\n¿Desea confiar permanentemente en el certificado?\nSe adjuntará automáticamente en su almacén de certificados de Windows",
                        "Advertencia", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
                    {
                        Keys.SavePublicKey(cert);
                        return true;
                    }
                    return false;
                }
            }
            catch (Exception e)
            {
                DoLog(e);
                return false;
            }
        }

        /// <summary>
        /// Reads xml from url, gets the public key and verifies the signature
        /// </summary>
        /// <param name="xml_not_verified"> xml to verify</param>
        /// <param name="toResponse">if it is from response</param>
        /// <param name="logout">If it is a logout try</param>
        /// <returns>if the signature was verified</returns>
        public static bool VerifySignature(XmlDocument xml_not_verified, bool toResponse, bool logout)
        {
            try
            {
                string publicKey;
                if (logout) publicKey = XMLData.ReadPublicKey(xml_not_verified,
                     "/saml2p:LogoutResponse/ds:Signature/ds:KeyInfo/ds:X509Data/ds:X509Certificate", true);

                else if (toResponse) publicKey = XMLData.ReadPublicKey(xml_not_verified,
                    "/saml2p:Response/ds:Signature/ds:KeyInfo/ds:X509Data/ds:X509Certificate", true);

                else publicKey = XMLData.ReadPublicKey(xml_not_verified,
                    "/md:EntityDescriptor/ds:Signature/ds:KeyInfo/ds:X509Data/ds:X509Certificate", false);

                X509Certificate2 pKey = new X509Certificate2(Encoding.ASCII.GetBytes(publicKey));

                //If you want to import a certificate that you haven't got in your key 
                //store and it have to be there, use the next method disabling the next comment

                //Keys.SavePublicKey(pKey);

                return (Sign.VerifyXmlFile(pKey, xml_not_verified));

            }
            catch (Exception e)
            {
                DoLog(e);
                return false;
            }
        }

        /// <summary>
        /// Example to make a log
        /// </summary>
        /// <param name="data">data for log</param>
        public static void DoLog(Exception data)
        {
            StreamWriter file = null;
            try
            {
                file = new StreamWriter(Conf.get_dir_log(), true);
                file.Write(GetDate());
                file.WriteLine(" " + data.Message);
                file.WriteLine(data.TargetSite);
                file.WriteLine(data.StackTrace);
                file.WriteLine(data.GetBaseException());
            }
            finally
            {
                file.Close();
            }
        }

        /// <summary>
        /// Creates a date
        /// </summary>
        /// <returns>date in string format</returns>
        public static string GetDate()
        {
            DateTime date = DateTime.UtcNow;

            return date.ToString("s");
        }

        /// <summary>
        /// Creates the IssueInstant metadata, the time when the link in the SP was clicked
        /// </summary>
        /// <returns></returns>
        public static string CreateIssueInstant()
        {
            DateTime date = DateTime.UtcNow;

            return date.ToString("s");
        }

        /// <summary>
        /// Creates a vaidUntil parameter for SP Metadata, 24 hours after the link in the SP was clicked
        /// </summary>
        /// <returns>string with a valid validUntil</returns>
        public static string CreateValidUntil()
        {
            DateTime date = DateTime.UtcNow;

            date = date.AddDays(1);

            return date.ToString("s");
        }

        class DefaultPasswordFinder : IPasswordFinder
        {
            private char[] password;

            public DefaultPasswordFinder(char[] password)
            {
                this.password = password;
            }

            public char[] GetPassword()
            {
                return (char[])password.Clone();
            }

        }
    }
}
