﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Net;
using System.Security.Cryptography.X509Certificates;
using System.IO;
using Newtonsoft.Json;
using log4net;

namespace Saml2UpdateCertCONSOLE
{
    public class UpdateCertWorker
    {
        private ILog _log;
        private const string CertificateDefinitionFilename = "certificates.json";
        private string _downloadUrl;
        private string _certPath;

        public ILog Log
        { 
            get { return _log; }
            set { _log = value; }
        }

        public UpdateCertWorker(string url, string certPath)
        {
            _downloadUrl = url;
            _certPath = certPath;
        }

        public void Run()
        {
            Log.Debug("Obteniendo XML de pasarela...");
            var xmlStr = GetXmlStrFromUrl(_downloadUrl);
            var downloadDate = DateTime.UtcNow.ToString("o", System.Globalization.CultureInfo.InvariantCulture);
            Log.Debug("Extrayendo información de certificados de XML...");
            List<ClaveCertificate> certificados = ExtractCertificatesFromXml(xmlStr, downloadDate);

            if (certificados.Count > 0)
            {
                Log.Debug("Limpiando directorio de certificados...");
                CleanCertificateDirectory();
                foreach (var certificado in certificados)
                {
                    Log.Debug($"Persistiendo certificado [{certificado.Name}]...");
                    PersistCertificate2FileSystem(certificado);
                }
                Log.Debug("Creando archivo de definición de certificados...");
                CreateCertificateDefinitionFile(certificados);
            }
        }

        private void CreateCertificateDefinitionFile(List<ClaveCertificate> certificados)
        {
            using (StreamWriter file = File.CreateText($"{_certPath}\\{CertificateDefinitionFilename}"))
            {
                JsonSerializer serializer = new JsonSerializer();
                serializer.Serialize(file, certificados);
            }
        }

        private void PersistCertificate2FileSystem(ClaveCertificate certificado)
        {
            X509Certificate2 cert = new X509Certificate2(Encoding.ASCII.GetBytes(certificado.PublicKey));
            File.WriteAllBytes($"{_certPath}\\{certificado.Name}.cer", cert.Export(X509ContentType.Cert));
            var useStr = certificado.Use == CertificateUse.Signing ? "firma" : "cifrado";
            Log.Info($"Certificado [{certificado.Name}] de {useStr} descargado");
        }

        private void CleanCertificateDirectory()
        {
            DirectoryInfo di = new DirectoryInfo(_certPath);
            foreach (FileInfo file in di.GetFiles()) { file.Delete(); }
        }

        private static List<ClaveCertificate> ExtractCertificatesFromXml(string xmlStr, string downloadDate)
        {
            List<ClaveCertificate> certificados = new List<ClaveCertificate>();

            XmlDocument doc = new XmlDocument();
            doc.LoadXml(xmlStr);
            XmlNamespaceManager namespaces = new XmlNamespaceManager(doc.NameTable);
            namespaces.AddNamespace("ds", "http://www.w3.org/2000/09/xmldsig#");
            namespaces.AddNamespace("md", "urn:oasis:names:tc:SAML:2.0:metadata");
            XmlNodeList nodes = doc.SelectNodes("/md:EntityDescriptor/md:IDPSSODescriptor/md:KeyDescriptor", namespaces);
            foreach (XmlNode node in nodes)
            {
                var certUse = node.Attributes["use"].Value == "signing" ? CertificateUse.Signing : CertificateUse.Encryption;
                var name = node.SelectSingleNode("ds:KeyInfo/ds:KeyName", namespaces).InnerText;
                var active = node.SelectSingleNode("ds:KeyInfo/ds:MgmtData", namespaces).InnerText.Equals("active", StringComparison.InvariantCultureIgnoreCase);
                var publicKey = node.SelectSingleNode("ds:KeyInfo/ds:X509Data/ds:X509Certificate", namespaces).InnerXml;

                certificados.Add(new ClaveCertificate(name, publicKey, certUse, active, downloadDate));
            }

            return certificados;
        }

        private static string GetXmlStrFromUrl(string url)
        {
            string ret = null;
            var uri = new Uri(url);
            if (uri.Scheme == "https") { ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls12; }
            using (WebClient client = new WebClient())
            {
                ret = client.DownloadString(url);
            }
            return ret;
        }
    }
}
